/*
								+--------------------------------+
								|                                |
								|    ***   File listener   ***   |
								|                                |
								|  Copyright  -tHE SWINe- 2008  |
								|                                |
								|          FileList.cpp          |
								|                                |
								+--------------------------------+
*/

#include <vector>
#include "../berLame_src/CallStack.h"
#include <stdio.h>
#include <algorithm>
#include "../berLame_src/Dir.h"
#include "../berLame_src/Integer.h"
#include "FileList.h"

/*
 *								=== CFileListener ===
 */

/*
 *	CFileListener::CFileListener(std::vector<TFileInfo*> &r_file_list)
 *		- default constructor; sets output file list (should be empty)
 *		- note the list will contain dynamically allocated data,
 *		  make sure to delete it, for example using Delete_FileList
 */
CFileListener::CFileListener(std::vector<TFileInfo*> &r_file_list)
	:m_r_file_list(r_file_list), m_n_total_size(0),
	m_b_size_valid(true), m_b_success(true)
{
	_ASSERTE(r_file_list.empty());
}

/*
 *	void CFileListener::operator ()(const TFileInfo &r_file_info)
 *		- executed by CDirTraversal, adds a new file to the list
 *		- this function may actually fail (in case there's not enough memory)
 *		  so converting this class to bool gives it's status
 *		- in case this fails, file list is guaranteed to be empty
 */
void CFileListener::operator ()(const TFileInfo &r_file_info)
{
	if(!m_b_success)
		return;
	// don't try anymore once failed

	if(!(m_b_success = ProcessFile(r_file_info)))
		Delete_FileList(m_r_file_list);
	// process file (ie. add it to the list)
}

/*
 *	uint64_t CFileListener::n_Total_Size() const
 *		- returns total size of all files in the list
 *		- note the size might be actually greater than maximum value of uint64_t
 *		  so there's function bool b_Size_Valid() which tells wheter value returned
 *		  by this function is valid
 */
uint64_t CFileListener::n_Total_Size() const
{
	return m_n_total_size;
}

/*
 *	bool CFileListener::b_Size_Valid() const
 *		- if value returned by n_Total_Size() is valid, returns true.
 *		  otherwise (overflow occured) returns false
 */
bool CFileListener::b_Size_Valid() const
{
	return m_b_size_valid;
}

/*
 *	operator CFileListener::bool() const
 *		- in case there was not enough memory for all files in the
 *		  file list, returns false. otherwise returns true.
 */
CFileListener::operator bool() const
{
	return m_b_success;
}

/*
 *	static inline void CFileListener::Delete_FileList(std::vector<TFileInfo*> &r_file_list)
 *		- utility cleanup function
 */
void CFileListener::Delete_FileList(std::vector<TFileInfo*> &r_file_list)
{
	std::for_each(r_file_list.begin(), r_file_list.end(), DeleteFileInfo);
	r_file_list.clear();
}

/*
 *	static inline void CFileListener::DeleteFileInfo(TFileInfo *p_file_info)
 *		- utility cleanup function
 */
inline void CFileListener::DeleteFileInfo(TFileInfo *p_file_info)
{
	delete p_file_info;
}

/*
 *	bool CFileListener::ProcessFile(const TFileInfo &r_file_info)
 *		- adds a new r_file_info file to the list
 *		- returns true on success, false on failure
 */
bool CFileListener::ProcessFile(const TFileInfo &r_file_info)
{
	if(r_file_info.b_directory)
		return true;
	// don't want directories

	if(!CheckRead(r_file_info.p_s_Filename())) {
		fprintf(stderr, "warning: error opening file \'%s\'"
			" (won't be included in the archive)\n", r_file_info.p_s_Filename());
		return true;
	}
	// try to open the file for reading

	if(m_r_file_list.size() == m_r_file_list.capacity()) {
		m_r_file_list.reserve((m_r_file_list.capacity())? m_r_file_list.capacity() * 2 : 1);
		if(m_r_file_list.size() == m_r_file_list.capacity())
			return false;
	}
	// make sure there's some space in the file list

	TFileInfo *p_file_info;
	if(!(p_file_info = new TFileInfo(r_file_info)))
		return false;
	if(!p_file_info->p_s_Filename()) {
		delete p_file_info;
		return false;
	}
	// make a copy of file info

	m_r_file_list.push_back(p_file_info);
	// put file into the list

	uint64_t n_size = (uint64_t(r_file_info.n_size_hi) << 32) | r_file_info.n_size_lo;
	if(m_n_total_size > UINT64_MAX - n_size)
		m_b_size_valid = false;
	m_n_total_size += n_size;
	// calculate total file size

	return true;
}

/*
 *	static bool CFileListener::CheckRead(const char *p_s_filename)
 *		- returns true in case file p_s_filename can be opened for reading,
 *		  otherwise returns false
 */
bool CFileListener::CheckRead(const char *p_s_filename)
{
	FILE *p_fr;
	if(p_fr = fopen(p_s_filename, "rb")) {
		fclose(p_fr);
		return true;
	}
	// try to open the file for reading

	return false;
}

/*
 *								=== ~CFileListener ===
 */
