/*
								+--------------------------------+
								|                                |
								| *** LamePak2 (un)pack tool *** |
								|                                |
								|  Copyright  -tHE SWINe- 2008  |
								|                                |
								|            Main.cpp            |
								|                                |
								+--------------------------------+
*/

#include <vector>
#include "../berLame_src/CallStack.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <limits.h>
#include <algorithm>
#include <string>
#include "../berLame_src/Dir.h"
#include "../berLame_src/Timer.h"
#include "../berLame_src/Compress.h"
#include "../berLame_src/LamePak2.h"
#include "Progress.h"
#include "FileList.h"
#include "../berLame_src/Thread.h"

#if defined(_MSC_VER) && !defined(__MWERKS__) && !defined(for)
#define for if(0) {} else for
#endif
// msvc 'for' scoping hack

#ifdef WIN32
static const char n_path_separator = '\\';
#else
static const char n_path_separator = '/';
#endif
// path separator for windows / linux

/*
 *	void PrintHelp()
 *		- prints helpscreen
 */
void PrintHelp()
{
	printf("pack: lamepak2 -d dir -o output.wad2 [-b|--block block-size] [-t|--thread thread-num]\n");
	printf("unpack: lamepak2 -i output.wad2 -d output-dir\n");
}

/*
 *	static void CreatePath(const std::string &r_s_path)
 *		- creates all directories in path r_s_path (aparts from last one which is filename)
 */
static void CreatePath(const std::string &r_s_path)
{
#ifdef WIN32
	off_t n_pos = 0, n_pos2;
	while((n_pos2 = r_s_path.find(n_path_separator, n_pos)) != std::string::npos) {
		CreateDirectory(r_s_path.substr(0, n_pos2).c_str(), 0);
		n_pos = n_pos2 + 1;
	}
	// have to create directory structure level by level
#else
	off_t n_pos;
	if((n_pos = r_s_path.rfind(n_path_separator)) != std::string::npos) {
		std::string s_cmd("mkdir -p ");
		s_cmd += r_s_path.substr(0, n_pos);
		system(s_cmd.c_str());
		// use shell command 'mkdir -p <directory>' (-p is for silence when directory
		// exists and for creating whole path (not just the top-most directory))
	}
	// can create whole directory structure using a single command
#endif
}

/*
 *	bool Pack(const char *p_s_dir, const char *p_s_output,
 *		unsigned int n_block_size, bool b_force_store, int n_thread_num)
 *		- pack some directory p_s_dir to pak file p_s_output
 *		- n_block_size is packer operating block size
 *		- setting b_force_store to true forces no compression
 *		- n_thread_num is number of packer threads for multi-cpu machines
 *		- returns true on success, false on failure
 */
bool Pack(const char *p_s_dir, const char *p_s_output,
	unsigned int n_block_size, bool b_force_store, int n_thread_num)
{
	CTimer timer;
	float f_start_time = timer.f_Time();
	// remember start time

	std::vector<TFileInfo*> file_list;
	CFileListener my_listener(file_list);
	if(!CDirTraversal::Traverse(p_s_dir, my_listener) || !my_listener) {
		_ASSERTE(file_list.empty()); // make sure no unallocated memory slips here
		fprintf(stderr, "error: error scanning directory structure\n");
		return false;
	}
	// get file list from directory structure

	if(file_list.empty()) {
		fprintf(stderr, "error: no files to pack\n");
		return false;
	}
	// we might have removed some files because they couldn't be read

	if(my_listener.b_Size_Valid()) {
		uint64_t n_total_size = my_listener.n_Total_Size();
		uint64_t n_long_block_num = n_total_size / n_block_size;
		if(n_total_size % n_block_size)
			++ n_long_block_num;
		// calc number of blocks

		int n_block_num = -1;
		if(n_long_block_num < INT_MAX)
			n_block_num = int(n_long_block_num);
		// convert number of blocks to integer

		printf("found %d files, total size: ", file_list.size());
		CMyProgressIndicator::PrintSize(n_total_size);
		printf(" (%d blocks)\n", n_block_num);
		// print size of all files and number of blocks
	} else
		fprintf(stderr, "error: files too big. size information not available\n");
	// print info on how many data do we have to pack

	timer.f_Time();
	// sample time so the integer tick counter doesn't overflow

	CPak2PackerUtil packer;
	// my packer

	if(!packer.Set_BlockSize(n_block_size))
		fprintf(stderr, "error: packer rejected block size setting, using default\n");
	if(!packer.Set_CompressionLevel((b_force_store)? 0 : 1))
		fprintf(stderr, "error: packer rejected compression level setting, using default\n");
	if(!packer.Set_Thread_Num(n_thread_num))
		fprintf(stderr, "error: packer rejected thread count setting, using default\n");
	CMyProgressIndicator progress_indicator(my_listener.n_Total_Size(),
		my_listener.b_Size_Valid(), timer, f_start_time);
	packer.Set_VerboseFunc(&progress_indicator);
	// set packer options

	if(packer.CreatePak(p_s_output, file_list, p_s_dir)) {
		CFileListener::Delete_FileList(file_list);
		// cleanup file list

		printf("\n");
		progress_indicator.PrintCompression();
		// tell us how much bytes we saved

		return true;
	}
	// create pak, print time

	CFileListener::Delete_FileList(file_list);
	// cleanup file list

	return false;
}

/*
 *	bool Unpack(const char *p_s_pak_file, const char *p_s_dir)
 *		- unpack some pak file p_s_pak_file to some directory p_s_dir
 *		- returns true on success, false on failure
 */
bool Unpack(const char *p_s_pak_file, const char *p_s_dir)
{
	CTimer timer;
	float f_start_time = timer.f_Time();

	CLamePak2 pak(p_s_pak_file);
	if(!pak.b_Status()) {
		fprintf(stderr, "error: \'%s\' can't be opened"
			" or has wrong format / version\n", p_s_pak_file);
		return false;
	}
	printf("opened pak v%d.%02d, %d files inside\n",
		pak.n_Pak_Version() / 100, pak.n_Pak_Version() % 100, pak.n_File_Num());
	// open pak, print some info

	std::string s_output_path(p_s_dir);
	if(s_output_path[s_output_path.length() - 1] != n_path_separator)
		s_output_path += n_path_separator;
	// make sure output path ends with backslash

	TBuffer t_file; // have it here so it doesn't realloc so offen
	const char n_pak_path_separator = pak.n_Path_Separator();
	for(int i = 0, n = pak.n_File_Num(); i < n; ++ i) {
		float f_time = timer.f_Time() - f_start_time;
		float f_remaining_time = f_time / (float(i + 1) / n) - f_time;
		printf("unpacking file %d, time: ", i + 1);
		CMyProgressIndicator::PrintTime(f_time);
		printf(", remaining: ");
		CMyProgressIndicator::PrintTime(f_remaining_time);
		printf(" \r");
		// show progress

		std::string s_path(pak.p_s_FileName(i));
		// get i-th filename

		if(n_pak_path_separator != n_path_separator) {
			off_t n_pos = 0;
			while((n_pos = s_path.find(n_pak_path_separator, n_pos)) != std::string::npos)
				s_path[n_pos] = n_path_separator;
		}
		// replace his path separator for mine

		std::string s_whole_path(s_output_path);
		s_whole_path += s_path;
		// concat output directory and file path

		if(!pak.UnpackFile(i, t_file)) {
			fprintf(stderr, "error: error unpacking file %d (\'%s\')\n", i, pak.p_s_FileName(i));
			return false;
		}
		FILE *p_fw;
		CreatePath(s_whole_path);
		if(!(p_fw = fopen(s_whole_path.c_str(), "wb"))) {
			fprintf(stderr, "error: can't open \'%s\'\n", s_whole_path.c_str());
			return false;
		}
		if(t_file.n_Size() && fwrite(t_file.p_Data(), t_file.n_Size(), 1, p_fw) != 1) {
			fprintf(stderr, "error: error writing file \'%s\'\n", s_whole_path.c_str());
			fclose(p_fw);
			return false;
		}
		fclose(p_fw);
		// unpack and store
	}
	// iterate trough files

	float f_time = timer.f_Time() - f_start_time;
	printf("\nit took ");
	CMyProgressIndicator::PrintTime(f_time);
	printf("\n");
	// and we're done

	return true;
}

int main(int n_arg_num, const char **p_arg_list)
{
	const char *p_s_dir = 0;
	const char *p_s_output = 0;
	const char *p_s_input = 0;
	int n_block_size = 1048576 * 4; // four megabyte blocks
	bool b_force_store = false;
	int n_thread_num = 1;

	for(int i = 1; i < n_arg_num; ++ i) {
		if(!strcmp(p_arg_list[i], "-h") || !strcmp(p_arg_list[i], "--help")) {
			PrintHelp();
			return 0;
		} else if(!strcmp(p_arg_list[i], "-s") || !strcmp(p_arg_list[i], "--store"))
			b_force_store = true;
		else if(i + 1 == n_arg_num) {
			fprintf(stderr, "error: argument %s is either unknown or "
				"needs its value to follow it\n", p_arg_list[i]);
			return -1;
		} else if(!strcmp(p_arg_list[i], "-d"))
			p_s_dir = p_arg_list[++ i];
		else if(!strcmp(p_arg_list[i], "-o"))
			p_s_output = p_arg_list[++ i];
		else if(!strcmp(p_arg_list[i], "-i"))
			p_s_input = p_arg_list[++ i];
		else if(!strcmp(p_arg_list[i], "-b") || !strcmp(p_arg_list[i], "--block"))
			n_block_size = atol(p_arg_list[++ i]);
		else if(!strcmp(p_arg_list[i], "-t") || !strcmp(p_arg_list[i], "--thread"))
			n_thread_num = atol(p_arg_list[++ i]);
		else {
			fprintf(stderr, "error: unknown argument %s\n", p_arg_list[i]);
			return -1;
		}
	}
	// parse args

	if(!p_s_dir) {
		fprintf(stderr, "error: no dir specified\n");
		return -1;
	}
	if(!p_s_output && !p_s_input) {
		fprintf(stderr, "error: no pak file specified\n");
		return -1;
	}
	if(p_s_output && p_s_input) {
		fprintf(stderr, "error: can't simultaneously pack and unpack\n");
		return -1;
	}
	// check args

	printf("=== LamePak2 v%d.%02d ===\n\n", CPak2PackerUtil::n_Version() / 100,
		CPak2PackerUtil::n_Version() % 100);
	// print some info

	if(p_s_output) {
		if(n_block_size <= 1024 || n_block_size > 100 * 1048576) {
			fprintf(stderr, "error: block size must be in range 1k to 100M\n");
			return -1;
		}
		// check block size

		if(CThread::n_CPU_Num() != n_thread_num) {
			fprintf(stderr, "warning: there are %d CPU's on this system, you're running %d threads\n",
				CThread::n_CPU_Num(), n_thread_num);
		}
		// print number of CPU's

		if(!Pack(p_s_dir, p_s_output, n_block_size, b_force_store, n_thread_num))
			return -1;
		// create pak from files in some directory
	} else /*if(p_s_input)*/ {
		if(b_force_store)
			fprintf(stderr, "warning: forcing store compression when unpacking has no effect\n");
		if(n_thread_num != 1)
			fprintf(stderr, "warning: unpacker is implemented as single-threaded\n");
		// detect invalid commandline

		if(!Unpack(p_s_input, p_s_dir))
			return -1;
		// unpack some pak
	}

	return 0;
}
